#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "tap.h"
#include "th.h"
#include "errtrap.h"
#include "vulcan.h"


static void
stdProps(pfacet)
    octObject *pfacet;
{
    octObject prop;
    
    ohCreatePropStr(pfacet,&prop, "VIEWTYPE", "PHYSICAL");
    ohCreatePropStr(pfacet,&prop, "EDITSTYLE", "PHYSICAL");
    ohCreatePropStr(pfacet,&prop, "CELLCLASS", "LEAF");
    ohCreatePropStr(pfacet,&prop, "CELLTYPE", "CONTACT");
}


/*
 *  Create a rectangular terminal frame, attach it to formal terminal 'term'
 *  and to layer 'layer'.  Note that the terminal is attached to Cell but its
 *  implementation is not.
 *
 *  inputs:	term - name of terminal
 *		layer - name of layer
 *		x1, y1, etc. - coordinates in internal units
 *  4/3/86
 */
static void
termFrame( oCellPtr, term, layer, x1, y1, x2, y2)
    octObject *oCellPtr;
    char *term, *layer;
    octCoord x1, y1, x2, y2;
{
    octObject oTerm;
    octObject oLayer;
    octObject oBox;

    if ( ohGetByTermName( oCellPtr, &oTerm, term ) != OCT_OK ) {
	octStatus s = ohCreateTerm( oCellPtr, &oTerm, term );
	OH_ASSERT( s );
    }


    OH_ASSERT(ohGetOrCreateLayer(oCellPtr, &oLayer, layer));
    OH_ASSERT(ohCreateBox( &oLayer, &oBox, x1,y1,x2,y2));
    OH_ASSERT(octAttach(&oTerm, &oBox));
}


static void makeInterface( oCellPtr ) 
    octObject *oCellPtr;
{
    octObject interfaceFacet;
    interfaceFacet = *oCellPtr;
    interfaceFacet.contents.facet.facet = "interface";
    
    if (octOpenFacet(&interfaceFacet) < OCT_OK)  {
	errRaise("prim", 1, "can't open  %s.\n", ohFormatName( &interfaceFacet ) );
    }
    
    vulcan( oCellPtr, &interfaceFacet );
    
    stdProps(&interfaceFacet);
    (void) octFlushFacet(&interfaceFacet);
}

static void checkDR( oCellPtr, layerName, width, height )
    octObject *oCellPtr;
    char *layerName;
    int width;
    int height;
{
    octObject layer;
    int minRule;

    ohGetOrCreateLayer( oCellPtr, &layer, layerName );
    
    minRule = tapGetMinWidth( &layer );
    
    if ( minRule == TAP_NO_RULE ) return;
    
    if ( width < minRule || height < minRule ) {
	errRaise( "prim", 1, "Width of %s cannot be less than %d", 
		 layerName, minRule );
    }
}


/*
 *  Copyright (c) 1986 by Jeffrey L. Burns.  All rights reserved.
 *
 *  Make a single-layer contact ("routing terminal").
 *
 *  4/3/86
 *  last modified 11/7/87
 */

void
primOneLayerContact( oCellPtr, layer, width, height)
    octObject *oCellPtr;
    char *layer;
    int width, height;
{
    octObject techProp;
    int halfWidth;		/*  (width * UNITSPERLAMBDA)/2  */
    int halfHeight;		/*  (height * UNITSPERLAMBDA)/2  */

    if (width <= 0)  {
	errRaise( "prim", 1, "width value %d should be positive.\n", width);
    }
    if (height <= 0)  {
	errRaise( "prim",1,"height value %d should be positive.\n", height);
    }

    halfWidth = width/2;
    halfHeight = height/2;

    checkDR( oCellPtr, layer, width, height );
    /*
     *  The single geometry is a terminal frame.
     */
    termFrame(oCellPtr, "t",layer,-halfWidth,-halfHeight,halfWidth,halfHeight);

    stdProps(oCellPtr);
    
    OH_ASSERT( octFlushFacet(oCellPtr) );

    makeInterface( oCellPtr );
}


/*
 *  Copyright (c) 1986 by Jeffrey L. Burns.  All rights reserved.
 *
 *  Make a rectangular four-layer contact (e.g., poly1-poly2-metal1-contact cut).
 *  The contact cut is explicit.  Geometries are centered about (0,0).
 *  Both contents and interface facets are made.
 *
 *  globals:	Cell - OCT_FACET being constructed
 *
 *  inputs:	cutLayer - name of the contact cut layer
 *		layer1 - name of one layer (e.g., diffusion)
 *		layer2 - name of the other layer (e.g., metal)
 *		width - width of geos on layer1 and layers
 *		height - similar to width
 *
 *  4/3/86
 *  last modified 10/10/1992 by Edoardo Charbon
 */

void
primFourLayerContact(oCellPtr, cutLayer, cutLayer2, layer1, layer2, width, height)
    octObject *oCellPtr;
    char *cutLayer, *cutLayer2, *layer1, *layer2; /* layer names */
    int width, height;			          /* dimensions in oct units */
{
    int halfWidth;
    int halfHeight;
    int cutPitch, cutPitch2;
    int extraXCuts, extraXCuts2, totalXCuts, totalXCuts2;
    int extraXSpace, extraXSpace2, xSpaceIncrement, xSpaceIncrement2;
    int xCoord, xCoord2;
    int extraXSpaceLoc, extraXSpaceLoc2, totalExtraXSpace, totalExtraXSpace2;
    int extraYCuts, extraYCuts2, totalYCuts, totalYCuts2;
    int extraYSpace, extraYSpace2, ySpaceIncrement, ySpaceIncrement2;
    int yCoord, yCoord2;
    int extraYSpaceLoc, extraYSpaceLoc2, totalExtraYSpace, totalExtraYSpace2;
    int i, j;
    int UNITSPERLAMBDA;
    int CUT_SIZE, CUT_SIZE2;
    int CUT_OVERLAP;
    int CUT_SPACING, CUT_SPACING2;
    
    octObject lyr1,lyr2,lyrCon,lyrCon2;
    
    ohGetOrCreateLayer( oCellPtr, &lyr1, layer1 );
    ohGetOrCreateLayer( oCellPtr, &lyr2, layer2 );
    ohGetOrCreateLayer( oCellPtr, &lyrCon, cutLayer );
    ohGetOrCreateLayer( oCellPtr, &lyrCon2, cutLayer2 );

    UNITSPERLAMBDA = thLambda();
    CUT_SIZE    = tapGetMaxWidth( &lyrCon );
    CUT_SIZE2    = tapGetMaxWidth( &lyrCon2 );
    CUT_OVERLAP = MAX(tapGetMinOverlap( &lyr1, &lyrCon ), 
		      tapGetMinOverlap( &lyr2, &lyrCon2 ));
    CUT_OVERLAP = MAX( CUT_OVERLAP, tapGetMinOverlap( &lyrCon2, &lyrCon ));
    CUT_SPACING = tapGetMinSpacing( &lyrCon, &lyrCon );
    CUT_SPACING2 = tapGetMinSpacing( &lyrCon2, &lyrCon2 );
    
    CUT_SIZE /= UNITSPERLAMBDA;
    CUT_SIZE2 /= UNITSPERLAMBDA;
    CUT_OVERLAP /= UNITSPERLAMBDA;
    CUT_SPACING /= UNITSPERLAMBDA;
    CUT_SPACING2 /= UNITSPERLAMBDA;
    
    width /= UNITSPERLAMBDA;
    height /= UNITSPERLAMBDA;

    if (width < 2 * CUT_OVERLAP + MAX(CUT_SIZE, CUT_SIZE2))  {
	width = 2 * CUT_OVERLAP + MAX(CUT_SIZE, CUT_SIZE2);
    }


    if (height < 2 * CUT_OVERLAP + MAX(CUT_SIZE, CUT_SIZE2)) {
	height = 2 * CUT_OVERLAP + MAX(CUT_SIZE, CUT_SIZE2);
    }

    halfWidth = width * UNITSPERLAMBDA / 2;
    halfHeight = height * UNITSPERLAMBDA / 2;
    cutPitch = CUT_SIZE + CUT_SPACING;
    cutPitch2 = CUT_SIZE2 + CUT_SPACING2;

    /*
     *  The number of extra cuts is found by subtracting off the distance
     *  needed for one cut CUT_SIZEx + 2*CUT_OVERLAPx, 
     *  and then dividing by the pitch (CUT_SIZEx + CUT_SPACINGx).
     */
    extraXCuts = (width - (2 * CUT_OVERLAP + CUT_SIZE)) / cutPitch;
    extraXCuts2 = (width - (2 * CUT_OVERLAP + CUT_SIZE2)) / cutPitch2;
    totalXCuts = extraXCuts + 1;
    totalXCuts2 = extraXCuts2 + 1;

    /*
     *  Any extra space is to be distributed as evenly as possible, so find the
     *  total extra space available and the extra space per cut-to-cut gap.
     */
    if (totalXCuts == 1)  {
	totalExtraXSpace = width - (2 * CUT_OVERLAP + CUT_SIZE);
    } else  {
	totalExtraXSpace = width -
	    		   (extraXCuts * cutPitch + 2 * CUT_OVERLAP + CUT_SIZE);
    }
    xSpaceIncrement = totalExtraXSpace / (totalXCuts + 1);
    extraXSpace = totalExtraXSpace - ((totalXCuts + 1) * xSpaceIncrement);

    if (totalXCuts2 == 1)  {
	totalExtraXSpace2 = width - (2 * CUT_OVERLAP + CUT_SIZE2);
    } else  {
	totalExtraXSpace2 = width -
	    		  (extraXCuts2 * cutPitch2 + 2 * CUT_OVERLAP + CUT_SIZE2);
    }
    xSpaceIncrement2 = totalExtraXSpace2 / (totalXCuts2 + 1);
    extraXSpace2 = totalExtraXSpace2 - ((totalXCuts2 + 1) * xSpaceIncrement2);
    
    /*
     *  If the xSpaceIncrement is an even number of lambda, then all of the
     *  extra space is distributed evenly.  If it isn't and if the number of
     *  spaces is odd, the extra is put in the middle.  If it isn't and if
     *  the number of spaces is even, then the extra is put on the ends.
     */
    if (extraXSpace != 0)  {
	if ((totalXCuts + 1) % 2  ==  0)  {
	    /*
	     *  Even number of spaces; put extra on ends (perhaps
	     *  asymmetrically).
	     */
	    extraXSpace = extraXSpace / 2;
	    extraXSpaceLoc = 1;
	} else  {
	    /*
	     *  Odd number.
	     */
	    extraXSpaceLoc = totalXCuts / 2 + 1;
	}	    
    } else  {
	extraXSpaceLoc = -1;
    }

    if (extraXSpace2 != 0)  {
	if ((totalXCuts2 + 1) % 2  ==  0)  {
	    /*
	     *  Even number of spaces; put extra on ends (perhaps
	     *  asymmetrically).
	     */
	    extraXSpace2 = extraXSpace2 / 2;
	    extraXSpaceLoc2 = 1;
	} else  {
	    /*
	     *  Odd number.
	     */
	    extraXSpaceLoc2 = totalXCuts2 / 2 + 1;
	}	    
    } else  {
	extraXSpaceLoc2 = -1;
    }

    /*
     *  Repeat the above calculations for the vertical dimension.
     */
    extraYCuts = (height - (2 * CUT_OVERLAP + CUT_SIZE)) / cutPitch;
    totalYCuts = extraYCuts + 1;

    extraYCuts2 = (height - (2 * CUT_OVERLAP + CUT_SIZE2)) / cutPitch2;
    totalYCuts2 = extraYCuts2 + 1;

    if (totalYCuts == 1)  {
	totalExtraYSpace = height - (2 * CUT_OVERLAP + CUT_SIZE);
    } else  {
	totalExtraYSpace = height -
	    		   (extraYCuts * cutPitch + 2 * CUT_OVERLAP + CUT_SIZE);
    }
    ySpaceIncrement = totalExtraYSpace / (totalYCuts + 1);
    extraYSpace = totalExtraYSpace - ((totalYCuts + 1) * ySpaceIncrement);
    
    if (totalYCuts2 == 1)  {
	totalExtraYSpace2 = height - (2 * CUT_OVERLAP + CUT_SIZE2);
    } else  {
	totalExtraYSpace2 = height -
	    		   (extraYCuts2 * cutPitch2 + 2 * CUT_OVERLAP +CUT_SIZE2);
    }
    ySpaceIncrement2 = totalExtraYSpace2 / (totalYCuts2 + 1);
    extraYSpace2 = totalExtraYSpace2 - ((totalYCuts2 + 1) * ySpaceIncrement2);
    

    if (extraYSpace != 0)  {
	if ((totalYCuts + 1) % 2  ==  0)  {
	    extraYSpace = extraYSpace / 2;
	    extraYSpaceLoc = 1;
	} else  {
	    extraYSpaceLoc = totalYCuts / 2 + 1;
	}	    
    } else  {
	extraYSpaceLoc = -1;
    }
 
    if (extraYSpace2 != 0)  {
	if ((totalYCuts2 + 1) % 2  ==  0)  {
	    extraYSpace2 = extraYSpace2 / 2;
	    extraYSpaceLoc2 = 1;
	} else  {
	    extraYSpaceLoc2 = totalYCuts2 / 2 + 1;
	}	    
    } else  {
	extraYSpaceLoc2 = -1;
    }

    /*
     *  Scale up to internal units as needed.
     */
    xSpaceIncrement *= UNITSPERLAMBDA;
    extraXSpace *= UNITSPERLAMBDA;
    ySpaceIncrement *= UNITSPERLAMBDA;
    extraYSpace *= UNITSPERLAMBDA;
    cutPitch *= UNITSPERLAMBDA;

    xSpaceIncrement2 *= UNITSPERLAMBDA;
    extraXSpace2 *= UNITSPERLAMBDA;
    ySpaceIncrement2 *= UNITSPERLAMBDA;
    extraYSpace2 *= UNITSPERLAMBDA;
    cutPitch2 *= UNITSPERLAMBDA;
    
    termFrame(oCellPtr, "t", layer1, -halfWidth, -halfHeight, halfWidth, halfHeight);
    termFrame(oCellPtr, "t", layer2, -halfWidth, -halfHeight, halfWidth, halfHeight);
	
    yCoord = -halfHeight + CUT_OVERLAP * UNITSPERLAMBDA + ySpaceIncrement;
    for (j=1; j<=totalYCuts; j++)  {
	if (j == extraYSpaceLoc)  {
	    yCoord += extraYSpace;	/* all extra space here */
	}
	xCoord = -halfWidth + CUT_OVERLAP * UNITSPERLAMBDA + xSpaceIncrement;
	for (i=1; i<=totalXCuts; i++)  {
	    if (i == extraXSpaceLoc)  {
		xCoord += extraXSpace;	/* all extra space here */
	    }
	    termFrame(oCellPtr, "t", cutLayer,  xCoord, yCoord,
		      xCoord+CUT_SIZE * UNITSPERLAMBDA,
		      yCoord+CUT_SIZE * UNITSPERLAMBDA);
	    xCoord = xCoord + cutPitch + xSpaceIncrement;
	}
	yCoord = yCoord + cutPitch + ySpaceIncrement;
    }

    yCoord2 = -halfHeight + CUT_OVERLAP * UNITSPERLAMBDA + ySpaceIncrement2;
    for (j=1; j<=totalYCuts2; j++)  {
	if (j == extraYSpaceLoc2)  {
	    yCoord2 += extraYSpace2;	/* all extra space here */
	}
	xCoord2 = -halfWidth + CUT_OVERLAP * UNITSPERLAMBDA + xSpaceIncrement2;
	for (i=1; i<=totalXCuts2; i++)  {
	    if (i == extraXSpaceLoc2)  {
		xCoord2 += extraXSpace2;	/* all extra space here */
	    }
	    termFrame(oCellPtr, "t", cutLayer2,  xCoord2, yCoord2,
		      xCoord2+CUT_SIZE2 * UNITSPERLAMBDA,
		      yCoord2+CUT_SIZE2 * UNITSPERLAMBDA);
	    xCoord2 = xCoord2 + cutPitch2 + xSpaceIncrement2;
	}
	yCoord2 = yCoord2 + cutPitch2 + ySpaceIncrement2;
    }

    stdProps(oCellPtr);
    OH_ASSERT(octFlushFacet(oCellPtr));

    /*
     *  Done with the contents; do the interface.
     */
    makeInterface( oCellPtr );
}


/*
 *  Copyright (c) 1986 by Jeffrey L. Burns.  All rights reserved.
 *
 *  Make a rectangular three-layer contact (e.g., poly-metal1-contact cut).
 *  The contact cut is explicit.  Geometries are centered about (0,0).
 *  Both contents and interface facets are made.
 *
 *  globals:	Cell - OCT_FACET being constructed
 *
 *  inputs:	cutLayer - name of the contact cut layer
 *		layer1 - name of one layer (e.g., diffusion)
 *		layer2 - name of the other layer (e.g., metal)
 *		width - width of geos on layer1 and layers
 *		height - similar to width
 *
 *  4/3/86
 *  last modified 11/7/87
 */

void
primThreeLayerContact(oCellPtr, cutLayer, layer1, layer2, width, height)
    octObject *oCellPtr;
    char *cutLayer, *layer1, *layer2;	/* layer names */
    int width, height;			/* dimensions in oct units */
{
    int halfWidth;
    int halfHeight;
    int cutPitch;
    int extraXCuts, totalXCuts, extraXSpace, xSpaceIncrement, xCoord;
    int extraXSpaceLoc, totalExtraXSpace;
    int extraYCuts, totalYCuts, extraYSpace, ySpaceIncrement, yCoord;
    int extraYSpaceLoc, totalExtraYSpace;
    int i, j;
    int UNITSPERLAMBDA;
    int CUT_SIZE   ;
    int CUT_OVERLAP;
    int CUT_SPACING;

    octObject lyr1,lyr2,lyrCon;
    
    ohGetOrCreateLayer( oCellPtr, &lyr1, layer1 );
    ohGetOrCreateLayer( oCellPtr, &lyr2, layer2 );
    ohGetOrCreateLayer( oCellPtr, &lyrCon, cutLayer );
    UNITSPERLAMBDA = thLambda();
    CUT_SIZE    = tapGetMaxWidth( &lyrCon );
    CUT_OVERLAP = tapGetMinOverlap( &lyr1, &lyrCon );
    CUT_OVERLAP = MAX( CUT_OVERLAP, tapGetMinOverlap( &lyr2, &lyrCon ));
    CUT_SPACING = tapGetMinSpacing( &lyrCon, &lyrCon );

    
    CUT_SIZE /= UNITSPERLAMBDA;
    CUT_OVERLAP /= UNITSPERLAMBDA;
    CUT_SPACING /= UNITSPERLAMBDA;

    width /= UNITSPERLAMBDA;
    height /= UNITSPERLAMBDA;

    if (width < 2 * CUT_OVERLAP + CUT_SIZE)  {
	width = 2 * CUT_OVERLAP + CUT_SIZE;
    }

    if (height < 2 * CUT_OVERLAP + CUT_SIZE)  {
	height = 2 * CUT_OVERLAP + CUT_SIZE;
    }

    halfWidth = width * UNITSPERLAMBDA / 2;
    halfHeight = height * UNITSPERLAMBDA / 2;

    cutPitch = CUT_SIZE + CUT_SPACING;

    /*
     *  The number of extra cuts is found by subtracting off the distance
     *  needed for one cut (CUT_SIZE + 2*CUT_OVERLAP), and then dividing by
     *  the cut pitch (CUT_SIZE + CUT_SPACING).
     */
    extraXCuts = (width - (2 * CUT_OVERLAP + CUT_SIZE)) / cutPitch;
    totalXCuts = extraXCuts + 1;

    /*
     *  Any extra space is to be distributed as evenly as possible, so find the
     *  total extra space available and the extra space per cut-to-cut gap.
     */
    if (totalXCuts == 1)  {
	totalExtraXSpace = width - (2 * CUT_OVERLAP + CUT_SIZE);
    } else  {
	totalExtraXSpace = width -
	    		   (extraXCuts * cutPitch + 2 * CUT_OVERLAP + CUT_SIZE);
    }
    xSpaceIncrement = totalExtraXSpace / (totalXCuts + 1);
    extraXSpace = totalExtraXSpace - ((totalXCuts + 1) * xSpaceIncrement);
    
    /*
     *  If the xSpaceIncrement is an even number of lambda, then all of the
     *  extra space is distributed evenly.  If it isn't and if the number of
     *  spaces is odd, the extra is put in the middle.  If it isn't and if
     *  the number of spaces is even, then the extra is put on the ends.
     */
    if (extraXSpace != 0)  {
	if ((totalXCuts + 1) % 2  ==  0)  {
	    /*
	     *  Even number of spaces; put extra on ends (perhaps
	     *  asymmetrically).
	     */
	    extraXSpace = extraXSpace / 2;
	    extraXSpaceLoc = 1;
	} else  {
	    /*
	     *  Odd number.
	     */
	    extraXSpaceLoc = totalXCuts / 2 + 1;
	}	    
    } else  {
	extraXSpaceLoc = -1;
    }

    /*
     *  Repeat the above calculations for the vertical dimension.
     */
    extraYCuts = (height - (2 * CUT_OVERLAP + CUT_SIZE)) / cutPitch;
    totalYCuts = extraYCuts + 1;

    if (totalYCuts == 1)  {
	totalExtraYSpace = height - (2 * CUT_OVERLAP + CUT_SIZE);
    } else  {
	totalExtraYSpace = height -
	    		   (extraYCuts * cutPitch + 2 * CUT_OVERLAP + CUT_SIZE);
    }
    ySpaceIncrement = totalExtraYSpace / (totalYCuts + 1);
    extraYSpace = totalExtraYSpace - ((totalYCuts + 1) * ySpaceIncrement);
    
    if (extraYSpace != 0)  {
	if ((totalYCuts + 1) % 2  ==  0)  {
	    extraYSpace = extraYSpace / 2;
	    extraYSpaceLoc = 1;
	} else  {
	    extraYSpaceLoc = totalYCuts / 2 + 1;
	}	    
    } else  {
	extraYSpaceLoc = -1;
    }
 
    /*
     *  Scale up to internal units as needed.
     */
    xSpaceIncrement *= UNITSPERLAMBDA;
    extraXSpace *= UNITSPERLAMBDA;
    ySpaceIncrement *= UNITSPERLAMBDA;
    extraYSpace *= UNITSPERLAMBDA;
    cutPitch *= UNITSPERLAMBDA;
    
    termFrame(oCellPtr, "t", layer1, -halfWidth, -halfHeight, halfWidth, halfHeight);
    termFrame(oCellPtr, "t", layer2, -halfWidth, -halfHeight, halfWidth, halfHeight);
	
    yCoord = -halfHeight + CUT_OVERLAP * UNITSPERLAMBDA + ySpaceIncrement;
    for (j=1; j<=totalYCuts; j++)  {
	if (j == extraYSpaceLoc)  {
	    yCoord += extraYSpace;	/* all extra space here */
	}
	xCoord = -halfWidth + CUT_OVERLAP * UNITSPERLAMBDA + xSpaceIncrement;
	for (i=1; i<=totalXCuts; i++)  {
	    if (i == extraXSpaceLoc)  {
		xCoord += extraXSpace;	/* all extra space here */
	    }
	    termFrame(oCellPtr, "t", cutLayer,  xCoord, yCoord,
		      xCoord+CUT_SIZE * UNITSPERLAMBDA,
		      yCoord+CUT_SIZE * UNITSPERLAMBDA);
	    xCoord = xCoord + cutPitch + xSpaceIncrement;
	}
	yCoord = yCoord + cutPitch + ySpaceIncrement;
    }

    stdProps(oCellPtr);
    OH_ASSERT(octFlushFacet(oCellPtr));

    /*
     *  Done with the contents; do the interface.
     */
    makeInterface( oCellPtr );
}

